import pandas as pd
import numpy as np
import torch
import torch.nn as nn
from torch.utils.data import TensorDataset, DataLoader, RandomSampler, SequentialSampler, Dataset
import transformers
from transformers import AutoModel, BertTokenizerFast, AdamW, BertModel, RobertaModel, AutoTokenizer, AutoConfig
from transformers import get_linear_schedule_with_warmup
from sklearn.metrics import matthews_corrcoef, mean_squared_error, median_absolute_error
from tqdm import tqdm
import logging
import psutil
import os
import gc
from sklearn.utils.class_weight import compute_class_weight
from e2eml.full_processing import postprocessing, cpu_processing_nlp
import random

# specify GPU
scaler = torch.cuda.amp.GradScaler()  # GPU
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')


class BERTDataSet(Dataset):
    def __init__(self, sentences, targets, tokenizer, max_length):
        self.sentences = sentences
        self.targets = targets
        self.tokenizer = tokenizer
        self.max_sen_length = max_length

    def __len__(self):
        return len(self.sentences)

    def __getitem__(self, idx):
        sentence = str(self.sentences[idx])
        target = self.targets[idx]
        inputs = self.tokenizer.encode_plus(
            sentence,
            None,
            add_special_tokens=True,
            max_length=int(self.max_sen_length*1.2),  # changed from static 300
            padding='max_length',
            return_token_type_ids=True,
            truncation=True
        )
        ids = inputs['input_ids']
        mask = inputs['attention_mask']
        token_type_ids = inputs["token_type_ids"]

        return {
            'ids': torch.tensor(ids, dtype=torch.long),
            'mask': torch.tensor(mask, dtype=torch.long),
            'token_type_ids': torch.tensor(token_type_ids, dtype=torch.long),
            'target': torch.tensor(target, dtype=torch.float)
        }


"""class BERTClass(torch.nn.Module):
    def __init__(self, transformer):
        super(BERTClass, self).__init__()
        self.bert = AutoModel.from_pretrained(transformer
                                              )
        config = AutoConfig.from_pretrained(transformer)
        config.update({"output_hidden_states": True,
                       "hidden_dropout_prob": 0.30,
                       #"layer_norm_eps": 1e-7
                       })

        self.roberta = AutoModel.from_pretrained(transformer, config=config)

        self.attention = nn.Sequential(
            nn.Linear(768, 512),
            nn.Tanh(),
            nn.Linear(512, 1),
            nn.Softmax(dim=1)
        )

        self.regressor = nn.Sequential(
            nn.Linear(768, 1)
        )

    def forward(self, input_ids, token_type_ids, attention_mask):
        roberta_output = self.roberta(input_ids=input_ids,
                                      token_type_ids=token_type_ids,
                                      attention_mask=attention_mask)

        # There are a total of 13 layers of hidden states.
        # 1 for the embedding layer, and 12 for the 12 Roberta layers.
        # We take the hidden states from the last Roberta layer.
        last_layer_hidden_states = roberta_output.hidden_states[-1]

        # The number of cells is MAX_LEN.
        # The size of the hidden state of each cell is 768 (for roberta-base).
        # In order to condense hidden states of all cells to a context vector,
        # we compute a weighted average of the hidden states of all cells.
        # We compute the weight of each cell, using the attention neural network.
        weights = self.attention(last_layer_hidden_states)

        # weights.shape is BATCH_SIZE x MAX_LEN x 1
        # last_layer_hidden_states.shape is BATCH_SIZE x MAX_LEN x 768
        # Now we compute context_vector as the weighted average.
        # context_vector.shape is BATCH_SIZE x 768
        context_vector = torch.sum(weights * last_layer_hidden_states, dim=1)

        # Now we reduce the context vector to the prediction score.
        return self.regressor(context_vector)"""


class NlpModel(postprocessing.FullPipeline, cpu_processing_nlp.NlpPreprocessing, BERTDataSet):
    """
    This class stores all model training and prediction methods for classification tasks.
    This class stores all pipeline relevant information (inherited from cpu preprocessing).
    The attribute "df_dict" always holds train and test as well as
    to predict data. The attribute "preprocess_decisions" stores encoders and other information generated during the
    model training. The attributes "predicted_classes" and "predicted_probs" store dictionaries (model names are dictionary keys)
    with predicted classes and probabilities (classification tasks) while "predicted_values" stores regression based
    predictions. The attribute "evaluation_scores" keeps track of model evaluation metrics (in dictionary format).
    :param datasource: Expects a Pandas dataframe (containing the target feature as a column)
    :param target_variable: Name of the target feature's column within the datasource dataframe.
    :param date_columns: Date columns can be passed as lists additionally for respective preprocessing. If not provided
    e2eml will try to detect datetime columns automatically. Date format is expected as YYYY-MM-DD anyway.
    :param categorical_columns: Categorical columns can be passed as lists additionally for respective preprocessing.
    If not provided e2eml will try to detect categorical columns automatically.
    :param nlp_columns: NLP columns expect a string declaring one text column.
    :param unique_identifier: A unique identifier (i.e. an ID column) can be passed as well to preserve this information
     for later processing.
    :param ml_task: Can be 'binary', 'multiclass' or 'regression'. On default will be determined automatically.
    :param preferred_training_mode: Must be 'cpu', if e2eml has been installed into an environment without LGBM and Xgboost on GPU.
    Can be set to 'gpu', if LGBM and Xgboost have been installed with GPU support. The default 'auto' will detect GPU support
    and optimize accordingly. (Default: 'auto')
    :param logging_file_path: Preferred location to save the log file. Will otherwise stored in the current folder.
    :param low_memory_mode: Adds a preprocessing feature to reduce dataframe memory footprint. Will lead to a loss in
    model performance. Will be extended by further memory savings features in future releases.
    However we highly recommend GPU usage to heavily decrease model training times.
    """
    def create_train_dataset(self):
        logging.info('Create NLP train dataset.')
        logging.info(f'RAM memory {psutil.virtual_memory()[2]} percent used.')
        X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
        tokenizer = self.preprocess_decisions[f"nlp_transformers"][f"transformer_tokenizer_{self.transformer_chosen}"]
        train_dataset = BERTDataSet(X_train[self.nlp_transformer_columns], Y_train, tokenizer, self.preprocess_decisions[f"nlp_transformers"][f"max_sentence_len"])
        return train_dataset

    def create_test_dataset(self):
        logging.info('Create NLP test dataset.')
        logging.info(f'RAM memory {psutil.virtual_memory()[2]} percent used.')
        X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
        tokenizer = self.preprocess_decisions[f"nlp_transformers"][f"transformer_tokenizer_{self.transformer_chosen}"]
        test_dataset = BERTDataSet(X_test[self.nlp_transformer_columns], Y_test, tokenizer, self.preprocess_decisions[f"nlp_transformers"][f"max_sentence_len"])
        return test_dataset

    def create_pred_dataset(self):
        logging.info('Create NLP prediction dataset.')
        logging.info(f'RAM memory {psutil.virtual_memory()[2]} percent used.')
        if self.prediction_mode:
            self.dataframe[self.target_variable] = 999  # creating dummy column
            dummy_target = self.dataframe[self.target_variable]
            self.dataframe.drop(self.target_variable, axis=1)
            tokenizer = self.preprocess_decisions[f"nlp_transformers"][f"transformer_tokenizer_{self.transformer_chosen}"]
            pred_dataset = BERTDataSet(self.dataframe[self.nlp_transformer_columns], dummy_target, tokenizer, self.preprocess_decisions[f"nlp_transformers"][f"max_sentence_len"])
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            tokenizer = self.preprocess_decisions[f"nlp_transformers"][f"transformer_tokenizer_{self.transformer_chosen}"]
            pred_dataset = BERTDataSet(X_test[self.nlp_transformer_columns], Y_test, tokenizer, self.preprocess_decisions[f"nlp_transformers"][f"max_sentence_len"])
        return pred_dataset

    def create_train_dataloader(self, train_batch_size=None, workers=None):
        if train_batch_size:
            pass
        else:
            train_batch_size = self.transformer_settings["train_batch_size"]

        if workers:
            pass
        else:
            workers = self.transformer_settings["num_workers"]
        logging.info('Create NLP train dataloader.')
        logging.info(f'RAM memory {psutil.virtual_memory()[2]} percent used.')
        train_dataset = self.create_train_dataset()
        train_dataloader = DataLoader(train_dataset, batch_size=train_batch_size, shuffle=True, num_workers=workers,
                                      pin_memory=True)
        return train_dataloader

    def create_test_dataloader(self, test_batch_size=None, workers=None):
        if test_batch_size:
            pass
        else:
            test_batch_size = self.transformer_settings["test_batch_size"]

        if workers:
            pass
        else:
            workers = self.transformer_settings["num_workers"]
        logging.info('Create NLP test dataloader.')
        logging.info(f'RAM memory {psutil.virtual_memory()[2]} percent used.')
        test_dataset = self.create_test_dataset()
        test_dataloader = DataLoader(test_dataset, batch_size=test_batch_size, shuffle=False, num_workers=workers,
                                     pin_memory=True)
        return test_dataloader

    def pred_dataloader(self, pred_batch_size=None, workers=None):
        if pred_batch_size:
            pass
        else:
            pred_batch_size = self.transformer_settings["pred_batch_size"]

        if workers:
            pass
        else:
            workers = self.transformer_settings["num_workers"]
        logging.info('Create NLP prediction dataloader.')
        logging.info(f'RAM memory {psutil.virtual_memory()[2]} percent used.')
        pred_dataset = self.create_pred_dataset()
        pred_dataloader = DataLoader(pred_dataset, batch_size=pred_batch_size, shuffle=False, num_workers=workers,
                                     pin_memory=True)
        return pred_dataloader

    def loss_fn(self, output, target):
        return torch.sqrt(nn.MSELoss()(output, target))

    # nn.MultiMarginLoss, #CrossEntropyLoss, #MSELoss

    def model_setup(self, epochs=None):
        if self.prediction_mode:
            pass
        else:
            logging.info('Define NLP model.')
            logging.info(f'RAM memory {psutil.virtual_memory()[2]} percent used.')
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            model = self.create_bert_regression_model(chosen_model=self.transformer_chosen)
            model.to(device)
            model.train()
            LR = 2e-5 #1e-3
            optimizer = AdamW(model.parameters(), LR, betas=(0.99, 0.998), weight_decay=1e-2)
            if epochs:
                pass
            else:
                epochs = self.transformer_settings["epochs"]
            epochs = epochs
            train_steps = int(len(X_train) / self.transformer_settings["train_batch_size"] * epochs)
            num_steps = int(train_steps * 0.1)
            scheduler = get_linear_schedule_with_warmup(optimizer, num_steps, train_steps)
            self.preprocess_decisions[f"nlp_transformers"][f"sheduler_{self.transformer_chosen}"] = scheduler
            return model, optimizer, train_steps, num_steps, scheduler

    def training(self, train_dataloader, model, optimizer, scheduler):
        logging.info('Start NLP training loop.')
        logging.info(f'RAM memory {psutil.virtual_memory()[2]} percent used.')
        model.train()
        allpreds = []
        alltargets = []
        self.reset_test_train_index()

        for a in train_dataloader:
            losses = []
            optimizer.zero_grad()

            with torch.cuda.amp.autocast():
                ids = a["ids"].to(device)
                mask = a["mask"].to(device)
                target = a["target"].to(device)
                token_type_ids = a["token_type_ids"].to(device)

                output = model(ids, mask)
                output = output[0].squeeze(-1)
                loss = self.loss_fn(output, target)

                # For scoring
                losses.append(loss.item() / len(output))
                allpreds.append(output.detach().cpu().numpy())
                alltargets.append(target.detach().squeeze(-1).cpu().numpy())

            scaler.scale(loss).backward()  # backwards of loss
            scaler.step(optimizer)  # Update optimizer
            scaler.update()  # scaler update
            scheduler.step()  # Update learning rate schedule

            # Combine dataloader minutes

        allpreds = np.concatenate(allpreds)
        alltargets = np.concatenate(alltargets)

        # I don't use loss, but I collect it
        losses = np.mean(losses)
        # Score with rmse
        train_rme_loss = np.sqrt(mean_squared_error(alltargets, allpreds))

        return losses, train_rme_loss

    def validating(self, valid_dataloader, model):
        logging.info('Start NLP validation loop.')
        logging.info(f'RAM memory {psutil.virtual_memory()[2]} percent used.')
        model.eval()
        allpreds = []
        alltargets = []

        for a in valid_dataloader:
            losses = []

            with torch.no_grad():
                ids = a["ids"].to(device)
                mask = a["mask"].to(device)
                target = a["target"].to(device)
                token_type_ids = a["token_type_ids"].to(device)

                output = model(ids, mask)
                output = output[0].squeeze(-1)
                loss = self.loss_fn(output, target)
                # For scoring
                losses.append(loss.item() / len(output))
                allpreds.append(output.detach().cpu().numpy())
                alltargets.append(target.detach().squeeze(-1).cpu().numpy())
                # Combine dataloader minutes

        allpreds = np.concatenate(allpreds)
        alltargets = np.concatenate(alltargets)

        # I don't use loss, but I collect it
        losses = np.mean(losses)
        # Score with rmse
        valid_rme_loss = np.sqrt(mean_squared_error(alltargets, allpreds))

        return allpreds, losses, valid_rme_loss

    def predicting(self, pred_dataloader, model, pathes):
        logging.info('Start NLP prediction loop.')
        logging.info(f'RAM memory {psutil.virtual_memory()[2]} percent used.')
        allpreds = []
        model_no = 0
        mode_cols = []
        for m_path in pathes:
            state = torch.load(m_path)
            model.load_state_dict(state["state_dict"])
            model.to(device)
            model.eval()
            preds = []
            allvalloss = 0
            with torch.no_grad():
                for a in pred_dataloader:
                    ids = a["ids"].to(device)
                    mask = a["mask"].to(device)
                    token_type_ids = a["token_type_ids"].to(device)
                    output = model(ids, mask, token_type_ids)
                    output = output[0].squeeze(-1)

                    preds.append(output.detach().cpu().numpy())

                preds = np.concatenate(preds)

                if self.prediction_mode:
                    self.dataframe[f"preds_model{model_no}"] = preds
                else:
                    X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
                    X_test[f"preds_model{model_no}"] = preds
                mode_cols.append(f"preds_model{model_no}")

                allpreds.append(preds)
                model_no += 1
            del state
            torch.cuda.empty_cache()
            _ = gc.collect()
        return allpreds, mode_cols

    def load_model_states(self, path=None):
        logging.info('Load model save states.')
        logging.info(f'RAM memory {psutil.virtual_memory()[2]} percent used.')
        if path:
            pass
        else:
            path = os.getcwd()
        if self.prediction_mode:
            pthes = [os.path.join(f"{path}/", s) for s in os.listdir(f"{path}/") if ".pth" in s]
            return pthes
        else:
            pthes = [os.path.join(f"{path}/", s) for s in os.listdir(f"{path}/") if ".pth" in s]
            return pthes

    def transformer_train(self):
        if self.prediction_mode:
            pass
        else:
            logging.info('Start NLP transformer training.')
            logging.info(f'RAM memory {psutil.virtual_memory()[2]} percent used.')
            self.reset_test_train_index()

            train_dataloader = self.create_train_dataloader()
            test_dataloader = self.create_test_dataloader()
            model, optimizer, train_steps, num_steps, scheduler = self.model_setup()
            scheduler = get_linear_schedule_with_warmup(optimizer, num_steps, train_steps)

            trainlosses = []
            vallosses = []
            bestscore = None
            trainscores = []
            validscores = []

            for epoch in tqdm(range(self.transformer_settings["epochs"])):
                print("---------------" + str(epoch) + "start-------------")
                trainloss, trainscore = self.training(train_dataloader, model, optimizer, scheduler)
                trainlosses.append(trainloss)
                trainscores.append(trainscore)
                print("trainscore is " + str(trainscore))
                preds, validloss, valscore = self.validating(test_dataloader, model)
                vallosses.append(validloss)
                validscores.append(valscore)

                print("valscore is " + str(valscore))
                if bestscore is None:
                    bestscore = valscore
                    print("Save first model")
                    state = {
                        'state_dict': model.state_dict(),
                        'optimizer_dict': optimizer.state_dict(),
                        "bestscore": bestscore
                    }
                    torch.save(state, "model0.pth")

                elif bestscore > valscore:
                    bestscore = valscore
                    print("found better point")
                    state = {
                        'state_dict': model.state_dict(),
                        'optimizer_dict': optimizer.state_dict(),
                        "bestscore": bestscore
                    }
                    torch.save(state, "model0.pth")
                else:
                    pass

            bestscores = []
            bestscores.append(bestscore)

            for fold in range(1, 5):

                self.reset_test_train_index()
                X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()

                # initializing the data
                train_dataloader = self.create_train_dataloader()
                test_dataloader = self.create_test_dataloader()

                model = self.create_bert_regression_model(chosen_model=self.transformer_chosen)
                model.to(device)
                LR = 2e-5 #1e-3
                optimizer = AdamW(model.parameters(), LR, betas=(0.99, 0.999), weight_decay=1e-2)
                train_steps = int(
                    len(X_train) / self.transformer_settings["train_batch_size"] * self.transformer_settings["epochs"])
                num_steps = int(train_steps * 0.1)
                scheduler = get_linear_schedule_with_warmup(optimizer, num_steps, train_steps)

                trainlosses = []
                vallosses = []
                bestscore = None
                trainscores = []
                validscores = []

                for epoch in tqdm(range(self.transformer_settings["epochs"])):
                    print("---------------" + str(epoch) + "start-------------")
                    trainloss, trainscore = self.training(train_dataloader, model, optimizer, scheduler)
                    trainlosses.append(trainloss)
                    trainscores.append(trainscore)

                    print("trainscore is " + str(trainscore))
                    preds, validloss, valscore = self.validating(test_dataloader, model)

                    vallosses.append(validloss)
                    validscores.append(valscore)
                    print("valscore is " + str(valscore))

                    if bestscore is None:
                        bestscore = valscore
                        print("Save first model")
                        state = {
                            'state_dict': model.state_dict(),
                            'optimizer_dict': optimizer.state_dict(),
                            "bestscore": bestscore
                        }
                        torch.save(state, "model" + str(fold) + ".pth")
                    elif bestscore > valscore:
                        bestscore = valscore
                        print("found better point")
                        state = {
                            'state_dict': model.state_dict(),
                            'optimizer_dict': optimizer.state_dict(),
                            "bestscore": bestscore
                        }
                        torch.save(state, "model" + str(fold) + ".pth")
                    else:
                        pass
                bestscores.append(bestscore)

            del model, optimizer, scheduler
            _ = gc.collect()

    def median_abs_error_eval(self, true_y, predicted):
        try:
            median_absolute_error_score = median_absolute_error(true_y, predicted)
        except Exception:
            median_absolute_error_score = 0
        return median_absolute_error_score

    def transformer_predict(self):
        logging.info('Start NLP transformer prediction.')
        logging.info(f'RAM memory {psutil.virtual_memory()[2]} percent used.')
        self.reset_test_train_index()
        model = self.create_bert_regression_model(chosen_model=self.transformer_chosen)
        pthes = self.load_model_states()
        print(pthes)
        pred_dataloader = self.pred_dataloader()
        allpreds, mode_cols = self.predicting(pred_dataloader, model, pthes)

        if self.prediction_mode:
            self.dataframe["transformers_median"] = self.dataframe[mode_cols].median(axis=1)
            self.dataframe["transformers_mean"] = self.dataframe[mode_cols].mean(axis=1)
            self.predicted_values['nlp_transformer'] = self.dataframe["transformers_mean"]
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()

            if self.transformer_settings["keep_best_model_only"]:
                # we check, if one savestate underperforms and delete him out
                scorings = []
                states = []
                for state in mode_cols:
                    state_score = self.median_abs_error_eval(Y_test, X_test[state])
                    #print(state_score)
                    scorings.append(state_score)
                    states.append(state)
                scorings_arr = np.array(scorings)
                scorings_mean = np.mean(scorings_arr)
                scorings_std = np.std(scorings_arr)
                #print(scorings_std)
                keep_state = scorings_arr < scorings_mean+scorings_std
                for index, state in enumerate(states):
                    os_string = state[-6:]
                    if keep_state.tolist()[index]:
                        pass
                    else:
                        states.remove(state)
                        X_test.drop(state, axis=1)
                        os.remove(f"{os_string}.pth")
            else:
                pass

            X_test["transformers_median"] = X_test[mode_cols].median(axis=1)
            X_test["transformers_mean"] = X_test[mode_cols].mean(axis=1)
            self.predicted_values['nlp_transformer'] = X_test["transformers_median"]


from e2eml.full_processing import cpu_preprocessing
from sklearn.decomposition import PCA
import re
import string
import spacy
import nltk
import ssl
from nltk.tokenize import word_tokenize
from nltk.corpus import wordnet, stopwords
from nltk import pos_tag
from textblob import TextBlob
import transformers
from transformers import AutoModel, AutoTokenizer, AutoConfig
from transformers import get_linear_schedule_with_warmup
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn import naive_bayes
from sklearn.linear_model import Ridge, ElasticNet
from sklearn.decomposition import TruncatedSVD
from imblearn.over_sampling import SMOTE
from vowpalwabbit.sklearn_vw import VWClassifier, VWRegressor
import lightgbm as lgb
import pandas as pd
import numpy as np
import random
import logging
import psutil
import gc

try:
    _create_unverified_https_context = ssl._create_unverified_context
except AttributeError:
    print("Created unverified SSL context to be able to download NLTK dictionaries.")
else:
    ssl._create_default_https_context = _create_unverified_https_context
nltk.download('punkt')
nltk.download('averaged_perceptron_tagger')
nltk.download('stopwords')
nltk.download('wordnet')
#stop_words = set(stopwords.words(self.blueprint_step_selection_nlp_transformers["synonym_language"]))

# TODO: Continue on NLP
"""
THIS PART IS UNFINISHED AND UNTESTED.
"""


class NlpPreprocessing(cpu_preprocessing.PreProcessing):
    """
    Preprocess a string.
    :parameter
        :param text: string - name of column containing text
        :param lst_stopwords: list - list of stopwords to remove
        :param flg_stemm: bool - whether stemming is to be applied
        :param flg_lemm: bool - whether lemmitisation is to be applied
    :return
        cleaned text
    """
    def utils_preprocess_text(self, text, flg_stemm=False, flg_lemm=True, lst_stopwords=None):
        logging.info('Start text cleaning.')
        logging.info(f'RAM memory {psutil.virtual_memory()[2]} percent used.')
        # clean (convert to lowercase and remove punctuations and   characters and then strip)
        text = re.sub(r'[^\w\s]', '', str(text).lower().strip())

        # Tokenize (convert from string to list)
        lst_text = text.split()    # remove Stopwords
        if lst_stopwords is not None:
            lst_text = [word for word in lst_text if word not in lst_stopwords]

        # Stemming (remove -ing, -ly, ...)
        if flg_stemm == True:
            logging.info('Start text stemming.')
            ps = nltk.stem.porter.PorterStemmer()
            lst_text = [ps.stem(word) for word in lst_text]

        # Lemmatisation (convert the word into root word)
        if flg_lemm == True:
            logging.info('Start text lemmatisation.')
            lem = nltk.stem.wordnet.WordNetLemmatizer()
            lst_text = [lem.lemmatize(word) for word in lst_text]

        # back to string from list
        text = " ".join(lst_text)
        return text

    def remove_url(self, text):
        logging.info('Start removing URLs.')
        url = re.compile(r'https?://\S+|www\.\S+')
        return url.sub(r'', str(text))

    def remove_punct(self, text):
        logging.info('Start removing punctuation.')
        table = str.maketrans('', '', string.punctuation)
        return text.translate(table)

    def remove_html(self, text):
        logging.info('Start removing HTML.')
        html = re.compile(r'<.*?>')
        return html.sub(r'', str(text))

    def remove_emoji(self, text):
        logging.info('Start removing emojis.')
        emoji_pattern = re.compile("["
                                   u"\U0001F600-\U0001F64F"  # emoticons
                                   u"\U0001F300-\U0001F5FF"  # symbols & pictographs
                                   u"\U0001F680-\U0001F6FF"  # transport & map symbols
                                   u"\U0001F1E0-\U0001F1FF"  # flags (iOS)
                                   u"\U00002702-\U000027B0"
                                   u"\U000024C2-\U0001F251"
                                   "]+", flags=re.UNICODE)
        return emoji_pattern.sub(r'', str(text))

    def decontraction(self, text):
        logging.info('Start decontraction.')
        text = re.sub(r"won\'t", " will not", text)
        text = re.sub(r"won\'t've", " will not have", text)
        text = re.sub(r"can\'t", " can not", text)
        text = re.sub(r"don\'t", " do not", text)

        text = re.sub(r"can\'t've", " can not have", text)
        text = re.sub(r"ma\'am", " madam", text)
        text = re.sub(r"let\'s", " let us", text)
        text = re.sub(r"ain\'t", " am not", text)
        text = re.sub(r"shan\'t", " shall not", text)
        text = re.sub(r"sha\n't", " shall not", text)
        text = re.sub(r"o\'clock", " of the clock", text)
        text = re.sub(r"y\'all", " you all", text)

        text = re.sub(r"n\'t", " not", text)
        text = re.sub(r"n\'t've", " not have", text)
        text = re.sub(r"\'re", " are", text)
        text = re.sub(r"\'s", " is", text)
        text = re.sub(r"\'d", " would", text)
        text = re.sub(r"\'d've", " would have", text)
        text = re.sub(r"\'ll", " will", text)
        text = re.sub(r"\'ll've", " will have", text)
        text = re.sub(r"\'t", " not", text)
        text = re.sub(r"\'ve", " have", text)
        text = re.sub(r"\'m", " am", text)
        text = re.sub(r"\'re", " are", text)
        return str(text)

    def seperate_alphanumeric(self, text):
        logging.info('Start seperating alphanumeric.')
        words = text
        words = re.findall(r"[^\W\d_]+|\d+", words)
        return " ".join(words)

    def cont_rep_char(self, text):
        tchr = text.group(0)

        if len(tchr) > 1:
            return tchr[0:2]

    def unique_char(self, rep, text):
        substitute = re.sub(r'(\w)\1+', rep, text)
        return str(substitute)

    def regex_clean_text_data(self):
        logging.info('Start text cleaning.')
        self.get_current_timestamp(task='Start text cleaning.')
        logging.info('Start text cleaning.')
        algorithm = 'regex_text'
        if self.prediction_mode:
            text_columns = []
            text_columns.append(self.nlp_transformer_columns)
            for text_col in text_columns:
                self.dataframe[text_col] = self.dataframe[text_col].apply(lambda  x: self.utils_preprocess_text(x))
                self.dataframe[text_col] = self.dataframe[text_col].apply(lambda x : self.remove_url(x))
                self.dataframe[text_col] = self.dataframe[text_col].apply(lambda x : self.remove_punct(x))
                self.dataframe[text_col] = self.dataframe[text_col].apply(lambda x : self.remove_emoji(x))
                self.dataframe[text_col] = self.dataframe[text_col].apply(lambda x : self.decontraction(x))
                self.dataframe[text_col] = self.dataframe[text_col].apply(lambda x : self.seperate_alphanumeric(x))
                self.dataframe[text_col] = self.dataframe[text_col].apply(lambda x : self.unique_char(self.cont_rep_char,x))
            logging.info('Finished text cleaning.')
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            text_columns = []
            text_columns.append(self.nlp_transformer_columns)

            for text_col in text_columns:
                try:
                    X_train[text_col] = X_train[text_col].apply(lambda  x: self.utils_preprocess_text(x))
                except (TypeError, AttributeError):
                    pass
                try:
                    X_train[text_col] = X_train[text_col].apply(lambda x : self.remove_url(x))
                except (TypeError, AttributeError):
                    pass
                try:
                    X_train[text_col] = X_train[text_col].apply(lambda x : self.remove_punct(x))
                except (TypeError, AttributeError):
                    pass
                try:
                    X_train[text_col] = X_train[text_col].apply(lambda x : self.remove_emoji(x))
                except (TypeError, AttributeError):
                    pass
                try:
                    X_train[text_col] = X_train[text_col].apply(lambda x : self.decontraction(x))
                except (TypeError, AttributeError):
                    pass
                try:
                    X_train[text_col] = X_train[text_col].apply(lambda x : self.seperate_alphanumeric(x))
                except (TypeError, AttributeError):
                    pass
                try:
                    X_train[text_col] = X_train[text_col].apply(lambda x : self.unique_char(self.cont_rep_char,x))
                except (TypeError, AttributeError):
                    pass

                try:
                    X_test[text_col] = X_test[text_col].apply(lambda  x: self.utils_preprocess_text(x))
                except (TypeError, AttributeError):
                    pass

                try:
                    X_test[text_col] = X_test[text_col].apply(lambda x : self.remove_url(x))
                except (TypeError, AttributeError):
                    pass
                try:
                    X_test[text_col] = X_test[text_col].apply(lambda x : self.remove_punct(x))
                except (TypeError, AttributeError):
                    pass
                try:
                    X_test[text_col] = X_test[text_col].apply(lambda x : self.remove_emoji(x))
                except (TypeError, AttributeError):
                    pass
                try:
                    X_test[text_col] = X_test[text_col].apply(lambda x : self.decontraction(x))
                except (TypeError, AttributeError):
                    pass
                try:
                    X_test[text_col] = X_test[text_col].apply(lambda x : self.seperate_alphanumeric(x))
                except (TypeError, AttributeError):
                    pass
                try:
                    X_test[text_col] = X_test[text_col].apply(lambda x : self.unique_char(self.cont_rep_char,x))
                except (TypeError, AttributeError):
                    pass
            logging.info('Finished text cleaning.')
            return self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)

    def textBlob_sentiment_polarity_score(self, text):
        # This polarity score is between -1 to 1
        polarity = TextBlob(text).sentiment.polarity
        return polarity

    def append_text_sentiment_score(self, text_columns=None):
        self.get_current_timestamp(task='Start text cleaning.')
        logging.info('Start sentiment polarity score.')
        algorithm = 'textblob_sentiment_score'
        if self.prediction_mode:
            text_columns = self.nlp_transformer_columns
            for text_col in [text_columns]:
                self.dataframe[f"{algorithm}_{text_col}"] = self.dataframe[text_col].apply(lambda x : self.textBlob_sentiment_polarity_score(x))
            logging.info('Finished text cleaning.')
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            text_columns = self.nlp_transformer_columns
            for text_col in [text_columns]:
                X_train[f"{algorithm}_{text_col}"] = X_train[text_col].apply(lambda x : self.textBlob_sentiment_polarity_score(x))
                X_test[f"{algorithm}_{text_col}"] = X_test[text_col].apply(lambda x : self.textBlob_sentiment_polarity_score(x))
            logging.info('Finished text cleaning.')
            return self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)

    def spacy_features(self, df: pd.DataFrame, text_column):
        """
        This function generates features using spacy en_core_wb_lg
        More information:
        https://www.kaggle.com/konradb/linear-baseline-with-cv
        https://www.kaggle.com/anaverageengineer/comlrp-baseline-for-complete-beginners

        """
        logging.info('Download spacy language package.')
        try:
            nlp = spacy.load('en_core_web_sm')
        except OSError:
            print('Downloading language model for the spaCy POS tagger\n'
                  "(don't worry, this will only happen once)")
            from spacy.cli import download
            download('en_core_web_sm')
            nlp = spacy.load('en_core_web_sm')
            # https://spacy.io/universe/project/spacy-transformers
            # https://spacy.io/models
            #!python -m spacy download en_core_web_trf
            #!pip install spacy-transformers
            # nlp = spacy.load('en_core_web_trf')
        # nlp = spacy.load('en_core_web_lg')
        with nlp.disable_pipes():
            vectors = [nlp(text).vector for text in df[text_column]]
        return vectors

    def get_spacy_col_names(self):
        logging.info('Get spacy column names.')
        names = list()
        for i in range(96):  # TODO: Why 96?
            names.append(f"spacy_{i}")
        return names

    def pos_tag_features(self, passage: str):
        """
        This function counts the number of times different parts of speech occur in an excerpt. POS tags are listed here:
        https://www.ling.upenn.edu/courses/Fall_2003/ling001/penn_treebank_pos.html
        """
        logging.info('Get POS tags.')
        pos_tags = ["CC", "CD", "DT", "EX", "FW", "IN", "JJ", "JJR", "JJS", "LS", "MD",
                    "NN", "NNS", "NNP", "NNPS", "PDT", "POS", "PRP", "RB", "RBR", "RBS", "RP", "TO", "UH",
                    "VB", "VBD", "VBG", "VBZ", "WDT", "WP", "WRB"]

        tags = pos_tag(word_tokenize(passage))
        tag_list = list()
        for tag in pos_tags:
            tag_list.append(len([i[0] for i in tags if i[1] == tag]))
        return tag_list

    def pos_tagging_pca_nlp_cols(self, df, text_cols, mode='fit', pca_pos_tags=True):
        if mode == 'transform':
            pass
        else:
            pos_columns = set([])

        if self.nlp_columns:
            text_cols = self.nlp_columns
        else:
            pass

        target_columns = ["CC", "CD", "DT", "EX", "FW", "IN", "JJ", "JJR", "JJS", "LS",
                          "MD",
                          "NN", "NNS", "NNP", "NNPS", "PDT", "POS", "PRP", "RB", "RBR",
                          "RBS", "RP", "TO", "UH",
                          "VB", "VBD", "VBG", "VBZ", "WDT", "WP", "WRB"]
        if mode == 'transform':
            for text_col in text_cols:
                temp_target_columns = [x + text_col for x in target_columns]
                df[text_col].fillna('None', inplace=True)
                spacy_df = pd.DataFrame(self.spacy_features(df, text_col), columns=self.get_spacy_col_names())
                temp_df = pd.merge(df, spacy_df, left_index=True, right_index=True, how='left')
                pos_df = pd.DataFrame(temp_df[text_col].apply(lambda p: self.pos_tag_features(p)).tolist(),
                                      columns=temp_target_columns)
                pos_df.fillna(0, inplace=True)
                if pca_pos_tags:
                    self.get_current_timestamp(task='PCA POS tags')
                    logging.info('Start to PCA POS tags.')
                    pca = self.preprocess_decisions[f"spacy_pos"][f"pos_pca_{text_col}"]
                    comps = pca.transform(pos_df)
                    pos_pca_cols = [f'POS PC-1 {text_col}', f'POS PC-2 {text_col}']
                    pos_df = pd.DataFrame(comps, columns=pos_pca_cols)
                    pos_df_pca = pos_df[pos_pca_cols]
                    df = pd.merge(df, pos_df_pca, left_index=True, right_index=True, how='left')
                else:
                    df = pd.merge(df, pos_df, left_index=True, right_index=True, how='left')
        elif mode == 'fit':

            # if self.nlp_columns
            nlp_columns = []
            for text_col in text_cols:
                try:
                    # do we have at least 3 words?
                    df[f'nof_words_{text_col}'] = df[text_col].apply(lambda s: len(s.split(' ')))
                    if df[f'nof_words_{text_col}'].max() >= 3:
                        temp_target_columns = [x + text_col for x in target_columns]
                        df[text_col].fillna('None', inplace=True)
                        spacy_df = pd.DataFrame(self.spacy_features(df, text_col), columns=self.get_spacy_col_names())
                        temp_df = pd.merge(df, spacy_df, left_index=True, right_index=True, how='left')
                        pos_df = pd.DataFrame(temp_df[text_col].apply(lambda p: self.pos_tag_features(p)).tolist(),
                                              columns=temp_target_columns)
                        pos_df.fillna(0, inplace=True)
                        for col in pos_df.columns:
                            pos_columns.add(col)
                        if pca_pos_tags:
                            self.get_current_timestamp(task='PCA POS tags')
                            logging.info('Start to PCA POS tags.')
                            pca = PCA(n_components=2)
                            comps = pca.fit_transform(pos_df)
                            self.preprocess_decisions[f"spacy_pos"][f"pos_pca_{text_col}"] = pca
                            pos_pca_cols = [f'POS PC-1 {text_col}', f'POS PC-2 {text_col}']
                            pos_df = pd.DataFrame(comps, columns=pos_pca_cols)
                            pos_df_pca = pos_df[pos_pca_cols]
                            df = pd.merge(df, pos_df_pca, left_index=True, right_index=True, how='left')
                        else:
                            df = pd.merge(df, pos_df, left_index=True, right_index=True, how='left')
                        nlp_columns.append(text_col)
                    else:
                        pass
                    df.drop(f'nof_words_{text_col}', axis=1, inplace=True)
                except AttributeError:
                    pass
            if pca_pos_tags:
                # get unique pos tag columns
                unique_pos_cols = list(set(pos_columns))
                self.preprocess_decisions[f"spacy_pos"]["pos_tagger_cols"] = unique_pos_cols

            if self.nlp_columns:
                pass
            else:
                # get unique original column names
                unique_nlp_cols = list(set(nlp_columns))
                self.nlp_columns = unique_nlp_cols
        try:
            del pca
            del spacy_df
            del temp_df
            del comps
            _ = gc.collect()
        except Exception:
            pass
        return df

    def pos_tagging_pca(self, pca_pos_tags=True):
        self.get_current_timestamp(task='Start Spacy, POS tagging')
        logging.info('Start spacy POS tagging loop.')
        algorithm = 'spacy_pos'
        if self.prediction_mode:
            text_columns = self.nlp_columns
            self.dataframe = self.pos_tagging_pca_nlp_cols(
                self.dataframe, text_columns, mode='transform', pca_pos_tags=pca_pos_tags)
            logging.info('Finished spacy POS tagging loop.')
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            if pca_pos_tags:
                self.preprocess_decisions[f"spacy_pos"] = {}
            else:
                pass
            if not self.nlp_columns:
                text_columns = X_train.select_dtypes(include=['object']).columns
            else:
                text_columns = self.nlp_columns
            X_train = self.pos_tagging_pca_nlp_cols(X_train,
                                                    text_columns, mode='fit',
                                                    pca_pos_tags=pca_pos_tags
                                                    )
            X_test = self.pos_tagging_pca_nlp_cols(X_test,
                                                   text_columns, mode='transform',
                                                   pca_pos_tags=pca_pos_tags
                                                   )
            logging.info('Finished spacy POS tagging loop.')
            return self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)

    def tfidf_pca(self, df, text_cols, mode='fit', pca_pos_tags=True, ngram_range=(1, 2)):
        if self.nlp_columns:
            text_cols = self.nlp_columns
        else:
            pass
        if mode == 'transform':
            for text_col in text_cols:
                df[text_col].fillna('None', inplace=True)
                try:
                    tfids = self.preprocess_decisions[f"tfidf_vectorizer"][f"tfidf_{text_col}"]
                    if pca_pos_tags:
                        vector = list(tfids.transform(df[text_col]).toarray())
                        self.get_current_timestamp(task='PCA POS tags')
                        logging.info('Start to PCA POS tags.')
                        pca = self.preprocess_decisions[f"tfidf_vectorizer"][f"tfidf_pca_{text_col}"]
                        comps = pca.transform(vector)
                        tfidf_pca_cols = [f'TFIDF PC-1 {text_col}', f'TFIDF PC-2 {text_col}']
                        pos_df = pd.DataFrame(comps, columns=tfidf_pca_cols)
                        tfidf_df_pca = pos_df[tfidf_pca_cols]
                        df = pd.merge(df, tfidf_df_pca, left_index=True, right_index=True, how='left')
                    else:
                        all_embeddings = tfids.transform(df[text_col]).toarray()
                        tfidf_df = pd.DataFrame(all_embeddings, columns=tfids.get_feature_names())
                        tfidf_df = tfidf_df.add_prefix("tfids_")
                        df = pd.concat([df, tfidf_df], axis=1)
                except KeyError:
                    pass
        elif mode == 'fit':
            if self.nlp_columns:
                text_cols = self.nlp_columns
            else:
                pass
            # if self.nlp_columnss
            nlp_columns = []
            for text_col in text_cols:
                try:
                    # do we have at least 1 word?
                    df[f'nof_words_{text_col}'] = df[text_col].apply(lambda s: len(s.split(' ')))
                    if df[f'nof_words_{text_col}'].max() >= 3:
                        df[text_col].fillna('None', inplace=True)
                        tfids = TfidfVectorizer(ngram_range=ngram_range,
                                                strip_accents='unicode',
                                                max_features=25000,
                                                max_df=0.85,
                                                min_df=3,
                                                norm='l2',
                                                sublinear_tf=True)
                        self.preprocess_decisions[f"tfidf_vectorizer"][f"tfidf_{text_col}"] = tfids
                        if pca_pos_tags:
                            vector = list(tfids.fit_transform(df[text_col]).toarray())
                            self.get_current_timestamp(task='PCA TfIDF matrix')
                            logging.info('Start to PCA TfIDF matrix.')
                            pca = PCA(n_components=2)
                            comps = pca.fit_transform(vector)
                            self.preprocess_decisions[f"tfidf_vectorizer"][f"tfidf_pca_{text_col}"] = pca
                            tfidf_pca_cols = [f'TFIDF PC-1 {text_col}', f'TFIDF PC-2 {text_col}']
                            pos_df = pd.DataFrame(comps, columns=tfidf_pca_cols)
                            tfidf_df_pca = pos_df[tfidf_pca_cols]
                            df = pd.merge(df, tfidf_df_pca, left_index=True, right_index=True, how='left')
                        else:
                            all_embeddings = tfids.fit_transform(df[text_col]).toarray()
                            tfidf_df = pd.DataFrame(all_embeddings, columns=tfids.get_feature_names())
                            tfidf_df = tfidf_df.add_prefix("tfids_")
                            df = pd.concat([df, tfidf_df], axis=1)
                        nlp_columns.append(text_col)
                    else:
                        pass
                    df.drop(f'nof_words_{text_col}', axis=1, inplace=True)
                except AttributeError:
                    pass
            if self.nlp_columns:
                pass
            else:
                # get unique original column names
                unique_nlp_cols = list(set(nlp_columns))
                self.nlp_columns = unique_nlp_cols
        try:
            del tfidf_df
            del all_embeddings
            del tfids
            del tfidf_df_pca
            _ = gc.collect()
        except Exception:
            pass
        return df

    def tfidf_vectorizer_to_pca(self, pca_pos_tags=True, ngram_range=(1, 3)):
        self.get_current_timestamp(task='Start TFIDF to PCA loop')
        logging.info('Start TFIDF to PCA loop.')
        algorithm = 'spacy_pos'
        if self.prediction_mode:
            text_columns = self.nlp_columns
            self.dataframe = self.tfidf_pca(
                self.dataframe, text_columns, mode='transform', pca_pos_tags=pca_pos_tags, ngram_range=ngram_range)
            logging.info('Finished TFIDF to PCA loop.')
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            if pca_pos_tags:
                self.preprocess_decisions[f"tfidf_vectorizer"] = {}
            else:
                pass
            if not self.nlp_columns:
                text_columns = X_train.select_dtypes(include=['object']).columns
            else:
                text_columns = self.nlp_columns
            X_train = self.tfidf_pca(X_train,
                                     text_columns, mode='fit',
                                     pca_pos_tags=pca_pos_tags,
                                     ngram_range=ngram_range
                                     )
            X_test = self.tfidf_pca(X_test,
                                    text_columns, mode='transform',
                                    pca_pos_tags=pca_pos_tags,
                                    ngram_range=ngram_range
                                    )
            logging.info('Finished TFIDF to PCA loop.')
            return self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)

    def tfidf_smote_text_columns(self):
        if self.prediction_mode:
            pass
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            tfids = TfidfVectorizer(
                                    strip_accents='unicode',
                                    max_features=25000,
                                    max_df=0.85,
                                    min_df=3,
                                    norm='l2',
                                    sublinear_tf=True)
            tfids.fit(X_train[self.nlp_transformer_columns])
            vectors = tfids.transform(X_train[self.nlp_transformer_columns])
            vectors_df = pd.DataFrame(vectors.todense(), columns=tfids.get_feature_names())
            smt = SMOTE(random_state=777, k_neighbors=1)
            X_SMOTE, y_SMOTE = smt.fit_sample(vectors_df, Y_train)
            X_train = pd.DataFrame(X_SMOTE.todense(), columns=tfids.get_feature_names())
            print(X_train)
            Y_train = y_SMOTE
            del vectors
            del vectors_df
            del smt
            _ = gc.collect()
            return self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)

    def get_synonyms(self, word):
        """
        Get synonyms of a word
        """
        synonyms = set()

        for syn in wordnet.synsets(word):
            for l in syn.lemmas():
                synonym = l.name().replace("_", " ").replace("-", " ").lower()
                synonym = "".join([char for char in synonym if char in ' qwertyuiopasdfghjklzxcvbnm'])
                synonyms.add(synonym)

        if word in synonyms:
            synonyms.remove(word)

        return list(synonyms)

    def synonym_replacement(self, words, n):
        words = words.split()
        new_words = words.copy()
        stop_words = nltk.corpus.stopwords.words(self.blueprint_step_selection_nlp_transformers["synonym_language"])
        random_word_list = list(set([word for word in words if word not in stop_words]))
        random.shuffle(random_word_list)
        num_replaced = 0

        for random_word in random_word_list:
            synonyms = self.get_synonyms(random_word)

            if len(synonyms) >= 1:
                synonym = random.choice(list(synonyms))
                new_words = [synonym if word == random_word else word for word in new_words]
                num_replaced += 1

            if num_replaced >= n: #only replace up to n words
                break
        sentence = ' '.join(new_words)
        return sentence

    def replace_synonyms_to_df_copy(self, words_to_replace=3, mode ='auto'):
        """
        The function copies the original dataframe and randomly exchanges words by their synonyms. It concats the original to the
        modified dataframe and returns the new training dataset.
        :param words_to_replace:
        :return:
        """
        if self.prediction_mode:
            pass
        else:
            X_train, X_test, Y_train, Y_test = self.unpack_test_train_dict()
            X_train[self.target_variable] = Y_train
            if mode == 'auto':
                sentence_length = X_train[self.nlp_transformer_columns].apply(lambda x: np.max([len(w) for w in x.split()]))
                words_to_replace = int(sentence_length.max()/15) # exchange fifteen percent of the words
            train_copy = X_train.copy()
            train_copy[self.nlp_transformer_columns] = train_copy[self.nlp_transformer_columns].apply(lambda x : self.synonym_replacement(x, words_to_replace))
            X_train = pd.concat([X_train, train_copy], ignore_index=True)
            X_train = X_train.reset_index(drop=True)
            Y_train = X_train[self.target_variable]
            X_train.drop(self.target_variable, axis=1)

            return self.wrap_test_train_to_dict(X_train, X_test, Y_train, Y_test)

    def create_bert_classification_model(self, chosen_model='bert-base-uncased'):
        logging.info('Start creating or loading transformer model for classification.')
        if not self.transformer_chosen:
            chosen_model = chosen_model
        config = AutoConfig.from_pretrained(self.transformer_chosen, num_labels=self.num_classes)
        model = transformers.AutoModelForSequenceClassification.from_pretrained(self.transformer_chosen, config=config)
        return model

    def create_bert_regression_model(self, chosen_model='bert-base-uncased'):
        logging.info('Start creating or loading transformer model for regression.')
        if not self.transformer_chosen:
            chosen_model = chosen_model
        config = AutoConfig.from_pretrained(self.transformer_chosen, num_labels=1)
        model = transformers.AutoModelForSequenceClassification.from_pretrained(self.transformer_chosen, config=config)
        return model

    def import_transformer_model_tokenizer(self, transformer_chosen=None):
        logging.info('Start importing transformer tokenizer.')
        if not transformer_chosen:
            transformer_chosen = 'bert-base-uncased'
        else:
            transformer_chosen = self.transformer_chosen

        if self.transformer_model_load_from_path:
            bert = AutoModel.from_pretrained(f"{self.transformer_model_load_from_path}",
                                             output_attentions=False,  # Whether the model returns attentions weights.
                                             output_hidden_states=False)
            tokenizer = transformers.AutoTokenizer.from_pretrained(f"{self.transformer_model_load_from_path}")
        else:
            if self.class_problem in ['binary', 'multiclass']:
                # import BERT-base pretrained model
                bert = self.create_bert_classification_model(transformer_chosen)
            elif self.class_problem == 'regression':
                bert = self.create_bert_regression_model(transformer_chosen)
            else:
                print("No correct ml_task defined during class instantiation.")
            # Load the BERT tokenizer
            tokenizer = AutoTokenizer.from_pretrained(transformer_chosen)
        if "nlp_transformers" in self.preprocess_decisions:
            pass
        else:
            self.preprocess_decisions[f"nlp_transformers"] = {}

        self.preprocess_decisions[f"nlp_transformers"][f"transformer_model_{transformer_chosen}"] = bert
        self.preprocess_decisions[f"nlp_transformers"][f"transformer_tokenizer_{transformer_chosen}"] = tokenizer



from e2eml.classification import classification_blueprints as cb
import pandas as pd
from sklearn.metrics import matthews_corrcoef
from sklearn.metrics import classification_report
from sklearn.utils import shuffle
import re
#import pytest


def load_titanic_data():
    """
    Load & preprocess Titanic dataset. The feature engineering simulates the business knowledge part.
    The code has been taken from:
    https://towardsdatascience.com/predicting-the-survival-of-titanic-passengers-30870ccc7e8
    :return: Several dataframes and series to be processed by blueprint.
    """
    data = pd.read_csv("titanic_train.csv")
    print('Create additional features and modify existing ones.')
    deck = {"A": 1, "B": 2, "C": 3, "D": 4, "E": 5, "F": 6, "G": 7, "U": 8}
    data['Cabin'] = data['Cabin'].fillna("U0")
    data['Deck'] = data['Cabin'].map(lambda x: re.compile("([a-zA-Z]+)").search(x).group())
    data['Deck'] = data['Deck'].map(deck)
    data['Deck'] = data['Deck'].fillna(0)
    data['Deck'] = data['Deck'].astype(int)

    data.loc[data['Fare'] <= 7.91, 'Fare'] = 0
    data.loc[(data['Fare'] > 7.91) & (data['Fare'] <= 14.454), 'Fare'] = 1
    data.loc[(data['Fare'] > 14.454) & (data['Fare'] <= 31), 'Fare'] = 2
    data.loc[(data['Fare'] > 31) & (data['Fare'] <= 99), 'Fare'] = 3
    data.loc[(data['Fare'] > 99) & (data['Fare'] <= 250), 'Fare'] = 4
    data.loc[data['Fare'] > 250, 'Fare'] = 5

    data['Age'].fillna(0, inplace=True)
    data['Age'] = data['Age'].astype(int)
    data.loc[data['Age'] <= 11, 'Age'] = 0
    data.loc[(data['Age'] > 11) & (data['Age'] <= 18), 'Age'] = 1
    data.loc[(data['Age'] > 18) & (data['Age'] <= 22), 'Age'] = 2
    data.loc[(data['Age'] > 22) & (data['Age'] <= 27), 'Age'] = 3
    data.loc[(data['Age'] > 27) & (data['Age'] <= 33), 'Age'] = 4
    data.loc[(data['Age'] > 33) & (data['Age'] <= 40), 'Age'] = 5
    data.loc[(data['Age'] > 40) & (data['Age'] <= 66), 'Age'] = 6
    data.loc[data['Age'] > 66, 'Age'] = 6

    titles = {"Mr": 1, "Miss": 2, "Mrs": 3, "Master": 4, "Rare": 5}
    # extract titles
    data['Title'] = data.Name.str.extract(' ([A-Za-z]+)\.', expand=False)
    # replace titles with a more common title or as Rare
    data['Title'] = data['Title'].replace(['Lady', 'Countess','Capt', 'Col','Don', 'Dr',
                                                 'Major', 'Rev', 'Sir', 'Jonkheer', 'Dona'], 'Rare')
    data['Title'] = data['Title'].replace('Mlle', 'Miss')
    data['Title'] = data['Title'].replace('Ms', 'Miss')
    data['Title'] = data['Title'].replace('Mme', 'Mrs')
    # convert titles into numbers
    data['Title'] = data['Title'].map(titles)
    # filling NaN with 0, to get safe
    data['Title'] = data['Title'].fillna(0)

    data['relatives'] = data['SibSp'] + data['Parch']
    data.loc[data['relatives'] > 0, 'not_alone'] = 0
    data.loc[data['relatives'] == 0, 'not_alone'] = 1
    data['not_alone'] = data['not_alone'].astype(int)

    data['Fare'] = data['Fare'].astype(int)
    data['Age_Class'] = data['Age']*data['Pclass']
    data['Fare_Per_Person'] = data['Fare']/(data['relatives']+1)
    data['Fare_Per_Person'] = data['Fare_Per_Person'].astype(int)
    print('Do dataframe splits.')
    test_df = data.head(800).copy()
    val_df = data.tail(91).copy()
    val_df_target = val_df["Survived"].copy()
    del val_df["Survived"]
    test_target = "Survived"
    test_categorical_cols = ["Pclass", "Name", 'Sex', 'PassengerId']
    return test_df, test_target, val_df, val_df_target, test_categorical_cols


def steel_fault_multiclass_data():
    data = pd.read_csv("faults.csv")
    data = shuffle(data)
    #data["leakage"] = data["target"]
    print(data.info())
    #from sklearn.preprocessing import LabelEncoder
    #le = LabelEncoder()
    #X = data.drop('target', axis=1)
    #Y = le.fit_transform(data['target'])

    #X['target'] = Y
    test_df = data.head(1500).copy()
    val_df = data.tail(441).copy()
    val_df_target = val_df["target"].copy()
    del val_df["target"]
    test_target = "target"
    test_categorical_cols = None
    return test_df, test_target, val_df, val_df_target, test_categorical_cols


def nlp_multiclass_data():
    data = pd.read_csv("Corona_NLP_train.csv", encoding='latin-1')
    test_df = data.head(100).copy()
    print(test_df[["OriginalTweet"]])
    val_df = data.tail(499).copy()
    val_df_target = val_df["Sentiment"].copy()
    del val_df["Sentiment"]
    test_target = "Sentiment"
    test_categorical_cols = ["Location", "OriginalTweet"]
    return test_df, test_target, val_df, val_df_target, test_categorical_cols


def test_ml_special_multiclass_full_processing_multimodel_max_voting(dataset='titanic'):
    if dataset == 'titanic':
        test_df, test_target, val_df, val_df_target, test_categorical_cols = load_titanic_data()
        titanic_auto_ml = cb.ClassificationBluePrint(datasource=test_df,
                                       target_variable=test_target,
                                       categorical_columns=test_categorical_cols,
                                                     preferred_training_mode='auto',
                                                     tune_mode='accurate')
        titanic_auto_ml.hyperparameter_tuning_sample_size = 1100

    elif dataset == 'synthetic_multiclass':
        test_df, test_target, val_df, val_df_target, test_categorical_cols = steel_fault_multiclass_data()
        titanic_auto_ml = cb.ClassificationBluePrint(datasource=test_df,
                                                     target_variable=test_target,
                                                     categorical_columns=test_categorical_cols)
    elif dataset == 'corona_tweet':
        test_df, test_target, val_df, val_df_target, test_categorical_cols = nlp_multiclass_data()
        titanic_auto_ml = cb.ClassificationBluePrint(datasource=test_df,
                                                     target_variable=test_target,
                                                     categorical_columns=test_categorical_cols,
                                                     preferred_training_mode='auto',
                                                     tune_mode='accurate',
                                                     nlp_transformer_columns='OriginalTweet'
                                                     )

    titanic_auto_ml.hyperparameter_tuning_rounds = {"xgboost": 10,
                                                    "lgbm": 8,
                                                    "tabnet": 3,
                                                    "ngboost": 10,
                                                    "sklearn_ensemble": 3,
                                                    "catboost": 10,
                                                    "ridge": 3,
                                                    "bruteforce_random": 10}
    titanic_auto_ml.special_blueprint_algorithms = {"ridge": False, #titanic, #synthetic_multiclass
                                                    "xgboost": False, #titanic
                                                    "ngboost": False, #titanic, #synthetic_multiclass
                                                    "lgbm": True, #titanic
                                                    "tabnet": False, #titanic
                                                    "vowpal_wabbit": True,
                                                    "sklearn_ensemble": False, #titanic, #synthetic_multiclass
                                                    "catboost": False
                                                    }
    titanic_auto_ml.blueprint_step_selection_non_nlp["synthetic_data_augmentation"] = False


    titanic_auto_ml.ml_special_multiclass_full_processing_multimodel_max_voting()
    titanic_auto_ml.ml_special_multiclass_full_processing_multimodel_max_voting(val_df)
    val_y_hat = titanic_auto_ml.predicted_classes['max_voting']

    algorithms = [alg for alg, value in titanic_auto_ml.special_blueprint_algorithms.items() if value]

    def get_matthews(algorithm):
        # Assess prediction quality on holdout data
        print(classification_report(pd.Series(val_df_target).astype(bool),
                                    titanic_auto_ml.predicted_classes[algorithm]))
        try:
            matthews = matthews_corrcoef(pd.Series(val_df_target).astype(bool),
                                         titanic_auto_ml.predicted_classes[algorithm])
        except Exception:
            print("Matthew failed.")
            matthews = 0
        print(matthews)

    for i in algorithms:
        print(f"---------Start evaluating {i}----------")
        get_matthews(i)
    finished = True
    assert finished == True


if __name__ == '__main__':
    test_ml_special_multiclass_full_processing_multimodel_max_voting('synthetic_multiclass')



from e2eml.regression import regression_blueprints as rb
import pandas as pd
from sklearn.metrics import mean_absolute_error
# track memory consumption in terminal: dmesg


def load_housingprices_data():
    """
    Load & preprocess Housing prices dataset. T
    :return: Several dataframes and series to be processed by blueprint.
    """
    data = pd.read_csv("housingprices_train.csv")

    def new_features(X):
        X['HasWoodDeck'] = (X['WoodDeckSF'] == 0) * 1
        X['HasOpenPorch'] = (X['OpenPorchSF'] == 0) * 1
        X['HasEnclosedPorch'] = (X['EnclosedPorch'] == 0) * 1
        X['Has3SsnPorch'] = (X['3SsnPorch'] == 0) * 1
        X['HasScreenPorch'] = (X['ScreenPorch'] == 0) * 1
        X['Total_Home_Quality'] = X['OverallQual'] + X['OverallCond']
        X['TotalSF'] = X['TotalBsmtSF'] + X['1stFlrSF'] + X['2ndFlrSF']
        X['TotalSquareFootage'] = (X['BsmtFinSF1'] + X['BsmtFinSF2'] + X['1stFlrSF'] + X['2ndFlrSF'])
        X['HasPool'] = X['PoolArea'].apply(lambda x: 1 if x > 0 else 0)
        X['Has2ndFloor'] = X['2ndFlrSF'].apply(lambda x: 1 if x > 0 else 0)
        X['HasGarage'] = X['GarageArea'].apply(lambda x: 1 if x > 0 else 0)
        X['HasBsmt'] = X['TotalBsmtSF'].apply(lambda x: 1 if x > 0 else 0)
        X['HasFireplace'] = X['Fireplaces'].apply(lambda x: 1 if x > 0 else 0)
        return X

    data = new_features(data)
    print('Do dataframe splits.')
    test_df = data.head(1000).copy()
    val_df = data.tail(460).copy()
    val_df_target = val_df["SalePrice"].copy()
    del val_df["SalePrice"]
    test_target = "SalePrice"
    test_categorical_cols = ["MSZoning", "Street", "Alley", "LotShape", "LotFrontage", "Street", "LandContour",
                             "Utilities", "LotConfig", "LandSlope", "Neighborhood", "Condition1", "Condition2",
                             "BldgType", "HouseStyle", "RoofStyle", "RoofMatl", "Exterior1st", "Exterior2nd",
                             "ExterQual", "ExterCond", "Foundation", "BsmtQual", "BsmtCond", "BsmtExposure",
                             "BsmtFinType1", "BsmtFinType2", "Heating", "HeatingQC", "Electrical", "KitchenQual",
                             "Functional", "FireplaceQU", "GarageType", "GarageYrBlt", "GarageFinish", "GarageQual",
                             "GarageCond", "PavedDrive", "PoolQC", "Fence", "MiscFeature", "SaleType", "SaleCondition"]
    return test_df, test_target, val_df, val_df_target, test_categorical_cols


def test_ml_special_regression_multiclass_full_processing_multimodel_avg_blender():
    test_df, test_target, val_df, val_df_target, test_categorical_cols = load_housingprices_data()
    titanic_auto_ml = rb.RegressionBluePrint(datasource=test_df,
                                   target_variable=test_target,
                                   categorical_columns=test_categorical_cols,
                                   preferred_training_mode='auto',
                                             tune_mode='accurate')
    titanic_auto_ml.hyperparameter_tuning_rounds = {"xgboost": 10,
                                                    "lgbm": 10,
                                                    "tabnet": 3,
                                                    "ngboost": 10,
                                                    "sklearn_ensemble": 3,
                                                    "catboost": 10,
                                                    "ridge": 3,
                                                    "bruteforce_random": 10,
                                                    "elasticnet": 10}
    titanic_auto_ml.special_blueprint_algorithms = {"ridge": True,
                                                    "xgboost": True,
                                                    "ngboost": True,
                                                    "lgbm": True,
                                                    "tabnet": False,
                                                    "vowpal_wabbit": True,
                                                    "sklearn_ensemble": True,
                                                    "catboost": True,
                                                    "elasticnet": True
                                                    }
    titanic_auto_ml.hyperparameter_tuning_sample_size = 800

    titanic_auto_ml.ml_special_regression_full_processing_multimodel_avg_blender()
    titanic_auto_ml.ml_special_regression_full_processing_multimodel_avg_blender(val_df)
    val_y_hat = titanic_auto_ml.predicted_values['blended_preds']
    mae = mean_absolute_error(val_df_target, val_y_hat)
    finished = True
    assert finished == True


def test_ml_bp10_train_test_regression_full_processing_linear_reg():
    test_df, test_target, val_df, val_df_target, test_categorical_cols = load_housingprices_data()
    titanic_auto_ml = rb.RegressionBluePrint(datasource=test_df,
                                             target_variable=test_target,
                                             categorical_columns=test_categorical_cols,
                                             preferred_training_mode='auto',
                                             tune_mode='accurate')
    titanic_auto_ml.ml_bp10_train_test_regression_full_processing_linear_reg()
    titanic_auto_ml.ml_bp10_train_test_regression_full_processing_linear_reg(val_df)
    val_y_hat = titanic_auto_ml.predicted_values['linear_regression']
    mae = mean_absolute_error(val_df_target, val_y_hat)
    finished = True
    assert finished == True

if __name__ == '__main__':
    test_ml_special_regression_multiclass_full_processing_multimodel_avg_blender()
